import abc

from protocol.payload.events import EventPost, OnlineStatus
from protocol.payload.functions import FuncCallRes, FuncCall
from protocol.payload.properties import PropPost, PropSetRes, PropGetRes, PropGet, PropSet
from protocol.payload.thingmodel import ThingModelPost, ThingModelPostRes
from protocol.topic.topic import TopicGet, TopicType
from util.JsonWrapper import JsonObj
from util.log import getLogger
from paho.mqtt import client as mqtt_client

from util.util import GetTimestamp, GetMessageId, jsonToStr, GetClientId, DoGZip

logger = getLogger("joylink")


class DeviceCallbackHandler(metaclass=abc.ABCMeta):
    """ 回调函数控制类 """

    @abc.abstractmethod
    def onConnect(self):
        """ 连接成功 回调函数 """
        pass

    @abc.abstractmethod
    def onThMdPostRes(self, objName: str, serviceKey: str, inRes: ThingModelPostRes):
        """ 物模型上报响应 回调函数 """
        pass

    @abc.abstractmethod
    def onPropSet(self, objName: str, serviceKey: str, inSet: PropSet):
        """ 属性设置 回调函数 """
        pass

    @abc.abstractmethod
    def onPropGet(self, objName: str, serviceKey: str, inGet: PropGet):
        """ 属性获取 回调函数 """
        pass

    @abc.abstractmethod
    def onFuncCall(self, objName: str, serviceKey: str, inCall: FuncCall):
        """ 方法调用 回调函数 """
        pass


# joylink 配置结构体定义
class JoylinkConfig(JsonObj):
    protocol: str = None
    hostname: str = None
    port: int = None
    deviceId: str = None
    username: str = None
    password: str = None
    cafile: str = None
    cert: str = None
    key: str = None
    insecure: bool = True
    gzip: bool = False


# joylink sdk 接口定义
class JoylinkSDK:
    mqtt: mqtt_client.Client = None
    version: str = None
    config: JoylinkConfig = None

    def __init__(self, mqtt: mqtt_client.Client, version: str, config: JoylinkConfig):
        self.mqtt = mqtt
        self.version = version
        self.config = config

    # sdk 欢迎函数
    def Welcome(self):
        """ sdk 欢迎函数 """
        logger.info(f"welcome to use joylink sdk 5.0 in python:{self.version}\n\n")

    # 初始化函数
    def Initialise(self):
        """ 初始化函数 """
        logger.info("Enter JoylinkSDk Initialise\n")
        # set sdk handle to devCallBack
        devCallBack.sdk = self
        # TODO
        logger.info("Leave JoylinkSDk Initialise\n\n")

    # 连接函数
    def Connect(self):
        """ 连接函数 """
        logger.info("Enter JoylinkSDk Connect\n")
        logger.info(f"Connecting to {self.config.protocol}://{self.config.hostname}:{self.config.port}")

        def on_connect_fail(client: mqtt_client.Client, userdata):
            logger.info("Failed to connect to mqtt!")

        # 订阅回调
        def on_subscribe(client, userdata, mid, granted_qos):
            logger.info(f"On Subscribed: mid = {mid} qos = {granted_qos}.")

        from core.message_callback import mqttCallbackOnConnect
        self.mqtt.on_connect_fail = on_connect_fail
        self.mqtt.on_connect = mqttCallbackOnConnect
        # self.mqtt.on_subscribe = on_subscribe

        self.mqtt.connect(self.config.hostname, self.config.port)
        self.mqtt.loop_start()

        logger.info("Connectting Successfuuly\n")
        logger.info("Leave JoylinkSDk Connect\n\n")

    # 设置设备回调 方法调用
    def SetDevCallBackHandler(self, handler: DeviceCallbackHandler):
        """ 设置设备回调 方法调用 """
        logger.info("Enter JoylinkSDk SetDevCallBackFuncCall\n")
        devCallBack.handler = handler
        logger.info("Leave JoylinkSDk SetDevCallBackFuncCall\n\n")

    # 设备上下线状态上报
    def OnlineStatusPost(self, objName: str, serviceKey: str, inPost: OnlineStatus):
        """ 设备上下线状态上报 """
        logger.info("Enter JoylinkSDk OnlineStatusPost\n")
        if inPost.messageId is None or len(inPost.messageId) == 0:
            inPost.messageId = GetMessageId("m")
        if inPost.timestamp is None or inPost.timestamp == 0:
            inPost.timestamp = GetTimestamp()
        # 获取mqtt客户端ID
        clientId = GetClientId(self.mqtt)

        # Marshal json
        msgPayload = jsonToStr(inPost)

        # 判断消息主题
        msgTopic: str
        if inPost.events == "connected":
            msgTopic = TopicGet(TopicType.EVT_ONLINE, objName, clientId, serviceKey)
        else:
            msgTopic = TopicGet(TopicType.EVT_OFFLINE, objName, clientId, serviceKey)
        self.mqtt.publish(topic=msgTopic, payload=DoGZip(msgPayload, self.config.gzip), qos=1, retain=False)
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk OnlineStatusPost\n\n")

    # 物模型上报函数
    def ThingModelPost(self, objName: str, serviceKey: str, inPost: ThingModelPost):
        """ 物模型上报函数 """
        logger.info("Enter JoylinkSDk ThingModelPost\n")
        if inPost.messageId is None or len(inPost.messageId) == 0:
            inPost.messageId = GetMessageId("m")
        if inPost.timestamp is None or inPost.timestamp == 0:
            inPost.timestamp = GetTimestamp()
        # 获取mqtt客户端ID
        clientId = GetClientId(self.mqtt)

        # Marshal json
        msgPayload = jsonToStr(inPost)
        msgTopic = TopicGet(TopicType.THINGMODEL_POST, objName, clientId, serviceKey)
        self.mqtt.publish(topic=msgTopic, payload=DoGZip(msgPayload, self.config.gzip), qos=1, retain=False)
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk ThingModelPost\n\n")

    # 属性上报
    def PropertiesPost(self, objName: str, serviceKey: str, inPost: PropPost):
        """ 属性上报 """
        logger.info("Enter JoylinkSDk PropertiesPost\n")
        if inPost.messageId is None or len(inPost.messageId) == 0:
            inPost.messageId = GetMessageId("m")
        if inPost.timestamp is None or inPost.timestamp == 0:
            inPost.timestamp = GetTimestamp()
        # 获取mqtt客户端ID
        clientId = GetClientId(self.mqtt)

        # Marshal json
        msgPayload = jsonToStr(inPost)
        msgTopic = TopicGet(TopicType.PROP_POST, objName, clientId, serviceKey)
        self.mqtt.publish(topic=msgTopic, payload=DoGZip(msgPayload, self.config.gzip), qos=1, retain=False)
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk PropertiesPost\n\n")

    # 属性设置响应
    def PropertiesSetResponse(self, objName: str, serviceKey: str, inRes: PropSetRes):
        """ 属性设置响应 """
        logger.info("Enter JoylinkSDk PropertiesSetResponse\n")
        if inRes.Timestamp is None or inRes.Timestamp == 0:
            inRes.Timestamp = GetTimestamp()

        # 获取mqtt客户端ID
        clientId = GetClientId(self.mqtt)

        # Marshal json
        msgPayload = jsonToStr(inRes)
        msgTopic = TopicGet(TopicType.PROP_SET_RES, objName, clientId, serviceKey)
        self.mqtt.publish(topic=msgTopic, payload=DoGZip(msgPayload, self.config.gzip), qos=1, retain=False)
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk PropertiesSetResponse\n\n")

    # 属性获取响应
    def PropertiesGetResponse(self, objName: str, serviceKey: str, inRes: PropGetRes):
        """ 属性获取响应 """
        logger.info("Enter JoylinkSDk PropertiesGetResponse\n")
        if inRes.Timestamp is None or inRes.Timestamp == 0:
            inRes.Timestamp = GetTimestamp()

        # 获取mqtt客户端ID
        clientId = GetClientId(self.mqtt)

        # Marshal json
        msgPayload = jsonToStr(inRes)
        msgTopic = TopicGet(TopicType.PROP_GET_RES, objName, clientId, serviceKey)
        self.mqtt.publish(topic=msgTopic, payload=DoGZip(msgPayload, self.config.gzip), qos=1, retain=False)
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk PropertiesGetResponse\n\n")

    # 方法调用响应
    def FunctionsCallResponse(self, objName: str, serviceKey: str, inRes: FuncCallRes):
        """ 方法调用响应 """
        logger.info("Enter JoylinkSDk FunctionsCallResponse\n")
        if inRes.Timestamp is None or inRes.Timestamp == 0:
            inRes.Timestamp = GetTimestamp()

        # 获取mqtt客户端ID
        clientId = GetClientId(self.mqtt)

        # Marshal json
        msgPayload = jsonToStr(inRes)
        msgTopic = TopicGet(TopicType.FUNC_CALL_RES, objName, clientId, serviceKey)
        self.mqtt.publish(topic=msgTopic, payload=DoGZip(msgPayload, self.config.gzip), qos=1, retain=False)
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk FunctionsCallResponse\n\n")

    # 事件上报
    def EventsPost(self, objName: str, serviceKey: str, inPost: EventPost):
        logger.info("Enter JoylinkSDk EventsPost\n")
        if inPost.messageId is None or len(inPost.messageId) == 0:
            inPost.messageId = GetMessageId("m")
        if inPost.timestamp is None or inPost.timestamp == 0:
            inPost.timestamp = GetTimestamp()

        # 获取mqtt客户端ID
        clientId = GetClientId(self.mqtt)

        # Marshal json
        msgPayload = jsonToStr(inPost)
        msgTopic = TopicGet(TopicType.EVT_POST, objName, clientId, serviceKey)
        self.mqtt.publish(topic=msgTopic, payload=DoGZip(msgPayload, self.config.gzip), qos=1, retain=False)
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk EventsPost\n\n")


# 创建新的sdk
def NewSDK(cfg: JoylinkConfig) -> JoylinkSDK:
    protocol = cfg.protocol
    broker = protocol + "://" + cfg.hostname + ":" + str(cfg.port)
    logger.info(broker)

    mqtt = mqtt_client.Client(cfg.deviceId)
    mqtt.username_pw_set(cfg.username, cfg.password)
    if protocol == "tls":
        # context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        mqtt.tls_set(ca_certs=cfg.cafile, certfile=cfg.cert, keyfile=cfg.key)
        mqtt.tls_insecure_set(cfg.insecure)

    s = JoylinkSDK(mqtt, "v1.0.0", cfg)
    return s


# 消息发送日志打印
def messagePublishPrint(t: str, p: str):
    logger.info(f"\033[1;33;60mMessage Publish\r\n{t}\r\n{p}\033[0m\r\n\r\n")


class DeviceCallback:
    """ 设备回调函数控制类 """

    def __init__(self):
        self._sdk = None
        self._handler = None

    @property
    def sdk(self):
        return self._sdk

    @sdk.setter
    def sdk(self, s: JoylinkSDK):
        self._sdk = s

    @property
    def handler(self):
        return self._handler

    @handler.setter
    def handler(self, h: DeviceCallbackHandler):
        self._handler = h


# 设备回调函数
devCallBack = DeviceCallback()
