from paho.mqtt.client import MQTTMessage, MQTT_ERR_SUCCESS
from core.joylink import devCallBack
from protocol.payload.functions import FuncCall
from protocol.payload.properties import PropSet, PropGet
from protocol.payload.thingmodel import ThingModelPostRes
from protocol.topic.topic import TopicGet, TopicParse, TopicType
from paho.mqtt import client as mqtt_client

from util.log import getLogger
from util.util import parseJson, GetClientId

logger = getLogger("message_callback")


def mqttCallbackOnConnect(client: mqtt_client.Client, userdata, flags, rc):
    """ MQTT 连接成功回调函数 """
    logger.info("Enter JoylinkSDk mqttCallbackOnConnect\n")
    if rc != 0:
        logger.error(f"Failed to connect, return code {rc}\n")
        return
    logger.info("Connected to MQTT Broker!")
    # 获取配置项
    clientId = GetClientId(client)
    # 订阅消息列表
    topics = [
        # 物模型上报响应
        TopicGet(TopicType.THINGMODEL_POST_RES, "+", clientId, ""),
        # 通用属性设置
        TopicGet(TopicType.PROP_SET, "+", clientId, ""),
        # 服务 属性设置
        TopicGet(TopicType.PROP_SET, "+", clientId, "+"),
        # 通用属性获取
        TopicGet(TopicType.PROP_GET, "+", clientId, ""),
        # 服务属性获取
        TopicGet(TopicType.PROP_GET, "+", clientId, "+"),
        # 通用方法调用
        TopicGet(TopicType.FUNC_CALL, "+", clientId, ""),
        # 服务方法调用
        TopicGet(TopicType.FUNC_CALL, "+", clientId, "+")
    ]
    for msgTopic in topics:
        logger.info(f"going to subscribe topic: {msgTopic}")
        # 订阅消息
        resTuple = client.subscribe(msgTopic, qos=1)
        if resTuple[0] == MQTT_ERR_SUCCESS:
            logger.info(f"subscribe topic success.{msgTopic}\n")
        else:
            logger.error(f"subscribe topic error.{msgTopic}\n")


    client.on_message = mqttCallbackOnMessage

    # 设备回调
    if devCallBack.handler is not None:
        devCallBack.handler.onConnect()
    logger.info("Leave JoylinkSDk mqttCallbackOnConnect\n\n")


# MQTT 消息处理回调函数
def mqttCallbackOnMessage(client: mqtt_client.Client, userdata, message: MQTTMessage):
    """ MQTT 消息处理回调函数 """
    logger.info("Enter JoylinkSDk mqttCallbackOnMessage\n")
    logger.info(f"joylink recv message\r\n{message.topic}\r\n{message.payload}\r\n")

    # 解析消息主题
    parsedTopic = TopicParse(message.topic)

    # 消息主题分发流程
    if parsedTopic.typeName == TopicType.TYPE_NAME_PROP:
        # 属性消息 处理
        if parsedTopic.command == "set":
            # 属性设置
            onProcMessagePropertiesSet(message, parsedTopic.objName, parsedTopic.serviceKey)
        elif parsedTopic.command == "get":
            # 属性获取
            onProcMessagePropertiesGet(message, parsedTopic.objName, parsedTopic.serviceKey)
        else:
            pass
    elif parsedTopic.typeName == TopicType.TYPE_NAME_THMD:
        # 物模型 消息处理
        if parsedTopic.command == "post" and parsedTopic.response == "response":
            # 物模型上报 响应
            onProcMessageThingModelPostRes(message, parsedTopic.objName, parsedTopic.serviceKey)
    elif parsedTopic.typeName == TopicType.TYPE_NAME_FUNC:
        # 方法调用 消息处理
        if parsedTopic.command == "call":
            # 属性设置
            onProcMessageFunctionsCall(message, parsedTopic.objName, parsedTopic.serviceKey)
    else:
        # TODO 其他消息处理
        pass

    logger.info("Leave JoylinkSDk mqttCallbackOnMessage\n\n")


# 物模型上报响应 消息处理函数
def onProcMessageThingModelPostRes(message: MQTTMessage, objName: str, serviceKey: str):
    """  物模型上报响应 消息处理函数 """
    logger.info("Enter JoylinkSDk onProcMessageThingModelPostRes\n")
    # 解析消息
    postRes = parseJson(message.payload.decode("UTF-8"), clazz=ThingModelPostRes)
    if devCallBack.handler is not None:
        devCallBack.handler.onThMdPostRes(devCallBack.sdk, objName, serviceKey, postRes)
    logger.info("Leave JoylinkSDk onProcMessageThingModelPostRes\n\n")


# 属性设置 消息处理函数
def onProcMessagePropertiesSet(message: MQTTMessage, objName: str, serviceKey: str):
    """ 属性设置 消息处理函数 """
    logger.info("Enter JoylinkSDk onProcMessagePropertiesSet\n")
    # 解析消息
    propSet = parseJson(message.payload.decode("UTF-8"), clazz=PropSet)
    if devCallBack.handler is not None:
        devCallBack.handler.onPropSet(objName, serviceKey, propSet)
    logger.info("Leave JoylinkSDk onProcMessagePropertiesSet\n\n")


# 属性获取 消息处理函数
def onProcMessagePropertiesGet(message: MQTTMessage, objName: str, serviceKey: str):
    """ 属性获取 消息处理函数 """
    logger.info("Enter JoylinkSDk onProcMessagePropertiesGet\n")
    # 解析消息
    propGet = parseJson(message.payload.decode("UTF-8"), clazz=PropGet)
    if devCallBack.handler is not None:
        devCallBack.handler.onPropGet(objName, serviceKey, propGet)
    logger.info("Leave JoylinkSDk onProcMessagePropertiesGet\n\n")


# 方法调用 消息处理函数
def onProcMessageFunctionsCall(message: MQTTMessage, objName: str, serviceKey: str):
    """ 方法调用 消息处理函数 """
    logger.info("Enter JoylinkSDk onProcMessageFunctionsCall\n")
    # 解析消息
    funcCall = parseJson(message.payload.decode("UTF-8"), clazz=FuncCall)
    if devCallBack.handler is not None:
        devCallBack.handler.onFuncCall(objName, serviceKey, funcCall)
    logger.info("Leave JoylinkSDk onProcMessageFunctionsCall\n\n")
